@@ -8,7 +8,7 @@ package Graph::Easy::As_svg;
 
 use vars qw/$VERSION/;
 
-$VERSION = '0.09';
+$VERSION = '0.10';
 
 use Graph::Easy;
 
@@ -17,10 +17,17 @@ use Graph::Easy;
 
 package Graph::Easy;
 
-sub EM () { 11; }
-
 use strict;
 
+sub EM
+  {
+  # return the height of one line in pixels, taking the font-size into account
+  my $self = shift;
+
+  # default is 14 pixels
+  $self->_font_size_in_pixels( 14 );
+  }
+
 my $devs = {
   'ah' => 
      "  <!-- open arrow head -->\n  <g id="
@@ -106,6 +113,96 @@ my $strokes = {
   'double-dash' => '4, 2',
   };
 
+sub _svg_use_def
+  {
+  # mark a certain def as used (to output it's definition later)
+  my ($self, $def_name) = @_;
+
+  $self->{_svg_defs}->{$def_name} = 1;
+  }
+
+sub text_styles_as_svg
+  {
+  my $self = shift;
+
+  my $style = '';
+  my $ts = $self->text_styles();
+
+  $style .= ' font-style="italic"' if $ts->{italic};
+  $style .= ' font-weight="bold"' if $ts->{bold};
+
+  if ($ts->{underline} || $ts->{none} || $ts->{overline} || $ts->{'line-through'})
+    {
+    # XXX TODO: HTML does seem to allow only one of them
+    my @s;
+    foreach my $k (qw/underline overline line-through none/)
+      {
+      push @s, $k if $ts->{$k};
+      }
+    my $s = join(' ', @s);
+    $style .= " text-decoration=\"$s\"" if $s;
+    }
+
+  # XXX TODO: this will needless include the font-size if set via
+  # "node { font-size: X }:
+
+  my $fs = $self->_font_size_in_pixels( 14 ); $fs = '' if $fs eq '14';
+
+  # XXX TODO:
+  # the 'style="font-size:XXpx"' is nec. for Batik 1.5 (Firefox and Opera also
+  # handle 'font-size="XXpx"'):
+
+  $style .= " style=\"font-size:${fs}px\"" if $fs;
+
+  $style;
+  }
+
+sub _svg_text
+  { 
+  # create a text via <text> at pos x,y, indented by "$indent"
+  my ($self, $label, $color, $em, $indent, $x, $y) = @_;
+
+  $label =~ s/\s*\\n\s*/\n/g;			# insert real newlines
+  my @lines = split/\n/,$label;			# split into lines
+
+  # We can't just join them togeter with 'x=".." dy="1em"' because Firefox 1.5
+  # doesn't support this (Batik does, tho). So calculate x and y on each tspan:
+  if (@lines > 1)
+    {
+    my $dy = $y;
+    $label = "$indent$indent\n<tspan y=\"$dy\">";
+    $dy += $em;
+    for my $i (0 .. @lines - 1)
+      {
+      my $join = "</tspan>\n$indent$indent<tspan x=\"$x\" y=\"$dy\">"; $dy += $em;
+      $label .= $lines[$i] . $join;
+      }
+    $label .= "</tspan>\n$indent";
+    }
+
+  my $fs; $fs = $self->text_styles_as_svg() if $label ne '';
+
+  my $link = $self->link();
+
+  # For an edge, the default stroke is black, but this will render a black
+  # outline around colored text. So disable the stroke with "non".
+  my $stroke = ''; $stroke = ' stroke="none"' if ref($self) =~ /Edge/;
+
+  my $svg = "$indent<text class=\"text\" x=\"$x\" y=\"$y\"$fs fill=\"$color\"$stroke>$label</text>\n";
+
+  if ($link ne '')
+    {
+    # although the title is already included on the outer shape, we need to
+    # add it to the link, too (for shape: none, and some user agents like
+    # FF 1.5 display the title only while outside the text-area)
+    my $title = $self->title(); $title = ' xlink:title="' . $title . '"' if $title ne '';
+    $svg = $indent . "<a xlink:href=\"$link\"$title>\n$indent" . $svg .
+           $indent . "</a>\n";
+    }
+
+  $svg . "\n"
+  }
+
 sub _as_svg
   {
   # convert the graph to SVG
@@ -127,20 +224,31 @@ EOSVG
 ;
     }
 
+  my $em = $self->EM();
+
   # XXX TODO: that should use the padding/margin attribute from the graph
-  my $xl = int(EM * 1); my $yl = int(EM * 1);
-  my $xr = int(EM * 1); my $yr = int(EM * 1);
+  my $xl = int($em); my $yl = int($em);
+  my $xr = int($em); my $yr = int($em);
 
   my $mx = $max_x + $xl + $xr;
   my $my = $max_y + $yl + $yr;
 
+  # we need both xmlns= and xmlns:xlink to make Firefix 1.5 happy :-(
   $txt .=
-     '<svg viewBox="0 0 ##MX## ##MY##" version="1.1" xmlns="http://www.w3.org/2000/svg">'
-    ."\n<!-- Generated by Graph::Easy v$Graph::Easy::VERSION" .
-	    " at " . scalar localtime() . " -->\n\n";
+#     '<svg viewBox="0 0 ##MX## ##MY##" width="##MX##" height="##MY##" version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">'
+     '<svg width="##MX##" height="##MY##" version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">'
+    ."\n<!-- Generated at " . scalar localtime() . " by:\n  " .
+     "Graph::Easy v$Graph::Easy::VERSION\n  Graph::Easy::As_svg v$Graph::Easy::As_svg::VERSION\n -->\n\n";
+
+  my $title = $self->title();
+
+  $txt .= " <title>$title</title>\n" if $title ne '';
 
   $txt .= " <defs>\n##devs##";
 
+  # clear used definitions
+  $self->{_svg_defs} = {};
+
   # which attributes must be output as what name:
   my $mutator = {
     background => 'fill',
@@ -148,33 +256,36 @@ EOSVG
     'color' => 'stroke',
     };
   my $skip = qr/^(
-   label|
-   linkbase|
+   arrow-style|
    (auto)?(link|title)|
-   nodeclass|
+   border-color|
    border-style|
-   padding.*|
-   margin.*|
+   border-width|
+   border|
+   color|
+   cols|
+   font-family|
+   label|
+   label-color|
+   linkbase|
    line-height|
    letter-spacing|
-   font-family|
-   border|
-   width|
-   color|
+   margin.*|
+   nodeclass|
+   padding.*|
+   rows|
+   size|
+   style
+   shape|
    title|
    text-align|
-   border-color|
-   shape|
-   rows|
-   cols|
-   size|
-   label-color|
-   arrow-style
+   text-style|
+   width|
    )\z/x;
 
   my $overlay = {
-    node => {
-      "text-anchor" => 'middle',
+    edge => {
+      "stroke" => 'black',
     },
   };
   # generate the class attributes first
@@ -198,13 +309,27 @@ EOSVG
  
   $txt .=" </defs>\n\n";
 
+  ###########################################################################
+  # prepare graph label output
+
+  my $label = $self->label();
+  my $lp = 'top';
+
+  if ($label ne '')
+    {
+    $lp = $self->attribute('label-pos') || 'top';
+
+    $my += $em * 2;
+    }
+
+  ###########################################################################
   # output the graph's background and border
 
   {
-  my $bg = $self->attribute('graph', 'fill'); $bg = 'white' unless defined $bg;
-  my $br = $self->attribute('graph', 'border-style'); $br = '' unless defined $br;
-  my $cl = $self->attribute('graph', 'border-color'); $cl = 'black' unless defined $cl;
-  my $bw = $self->attribute('graph', 'border-width'); $bw = '1' unless defined $bw;
+  my $bg = $self->attribute('fill'); $bg = 'white' unless defined $bg; 
+  my $br = $self->attribute('border-style'); $br = '' unless defined $br;
+  my $cl = $self->attribute('border-color'); $cl = 'black' unless defined $cl;
+  my $bw = $self->attribute('border-width'); $bw = '1' unless defined $bw;
 
   if ($bg ne 'white' || $br ne '' || $cl ne 'black')
     {
@@ -216,14 +341,29 @@ EOSVG
 
     # Provide some padding arund the graph to avoid that the border sticks
     # very close to the edge
-    $xl += EM * 1;
-    $yl += EM * 1;
+    $xl += $em;
+    $yl += $em;
 
-    $mx += EM * 2;
-    $my += EM * 2;
+    $mx += $em * 2;
+    $my += $em * 2;
     }
   }
 
+  ###########################################################################
+  # adjust space for the graph label and output the label
+
+  if ($label ne '')
+    {
+    my $y = $yl + $em; $y = $my - 2 * $em if $lp eq 'bottom';
+
+    $txt .= "<!-- graph label -->\n" .
+            Graph::Easy::Node::_svg_text($self, $label,
+		$self->attribute('color') || 'black', $em, '', $mx / 2, $y);
+
+    # push content down if label is at top
+    $yl += $em * 2 if $lp eq 'top';
+    }
+
   # XXX: output cells that belong to one edge/node in order, e.g.
 #  for my $n (@nodes)
 #     {
@@ -254,9 +394,7 @@ EOSVG
 
   # include the used definitions into <devs>
   my $d = '';
-  #for my $key (keys %$used)
-  for my $key (qw/ah ahc ahf circle square dot cross star diamond
-                  d-circle d-square d-diamond/)
+  for my $key (keys %{$self->{_svg_defs}})
     {
     $d .= $devs->{$key};
     }
@@ -265,8 +403,11 @@ EOSVG
   $txt =~ s/##MX##/$mx/;
   $txt =~ s/##MY##/$my/;
 
-  $txt . "</svg>";			# finish
+  $txt .= "</svg>";			# finish
 
+  $txt .= "\n" if $options->{standalone};
+
+  $txt;
   }
 
 =pod
@@ -291,6 +432,14 @@ Graph::Easy::As_svg - Output a Graph::Easy as Scalable Vector Graphics (SVG)
 C<Graph::Easy::As_svg> contains just the code for converting a
 L<Graph::Easy|Graph::Easy> object to a SVG text.
 
+X<graph::easy>
+X<graph>
+X<drawing>
+X<svg>
+X<scalable>
+X<vector>
+X<grafics>
+
 =head1 EXPORT
 
 Exports nothing.
@@ -305,6 +454,8 @@ Copyright (C) 2004 - 2005 by Tels L<http://bloodgate.com>
 
 See the LICENSE file for information.
 
+x<tels>
+
 =cut
 
 package Graph::Easy::Node;
@@ -316,7 +467,7 @@ sub _svg_background
 
   my $bg = $self->background();
 
-  if ($bg ne 'inherit')
+  if ($bg ne 'inherit' && $bg ne '')
     {
     my $w = $self->{w};
     my $h = $self->{h};
@@ -325,6 +476,13 @@ sub _svg_background
 
   }
 
+BEGIN
+  {
+  *EM = \&Graph::Easy::EM;
+  *text_styles_as_svg = \&Graph::Easy::text_styles_as_svg;
+  *_svg_text = \&Graph::Easy::_svg_text;
+  }
+
 sub as_svg
   {
   # output a node as SVG
@@ -332,7 +490,7 @@ sub as_svg
 
   my $name = $self->{att}->{label}; $name = $self->{name} unless defined $name;
 
-  my $em = Graph::Easy::EM;		# multiplication factor chars * em = units (pixels)
+  my $em = $self->EM();		# multiplication factor chars * em = units (pixels)
 
   # the attributes of the element we will finally output
   my $att = $self->_svg_attributes($x,$y);
@@ -375,41 +533,51 @@ sub as_svg
     # center a square point-node
     $yt -= 5 if $s =~ 'square';
     $xt -= 5 if $s =~ 'square';
+
+    $self->{graph}->_svg_use_def($s);
+
     $svg .= "$indent<use$att_txt xlink:href=\"#$s\" x=\"$xt\" y=\"$yt\"/>\n\n";
     }
   else
     {
     my $att_txt = $self->_svg_attributes_as_txt($att);
 
-    $svg .= "$indent<$shape$att_txt />\n";
-
-    if ($bs =~ /^double/)
+    if ($shape ne 'none')
       {
-      my $att = $self->_svg_attributes($x,$y, 3);
-
-      my $shape = $att->{shape};				# rect, circle etc
-      delete $att->{shape};
-
-      my $att_txt = $self->_svg_attributes_as_txt( $att );
       $svg .= "$indent<$shape$att_txt />\n";
+
+      # if border-style is double, do it again, sam.
+      if ($bs =~ /^double/)
+        {
+        my $att = $self->_svg_attributes($x,$y, 3);
+
+        my $shape = $att->{shape};				# rect, circle etc
+        delete $att->{shape};
+
+        my $att_txt = $self->_svg_attributes_as_txt( $att );
+        $svg .= "$indent<$shape$att_txt />\n";
+        }
       }
 
     ###########################################################################
     # include the label/name/text
 
+    my $xt = int($x + $em * 0.05 + $self->{w} / 2);
+
     my $label = $self->label();
 
-    $label =~ s/\\n/\n/g;				# insert real newlines
+    # count lines
+    my $lines = -1; $label =~ s/\\n/ $lines++; '\\n' /eg; 
 
-    my $xt = int($x + $em * 0.05 + $self->{w} / 2);
-    my $yt = int($y + $self->{h} / 2 + $em / 2 - $em * 0.1);
-
-    $yt += $em if $s =~ /^(triangle|trapezium)\z/;
-    $yt -= $em if $s =~ /^inv(triangle|trapezium)\z/;
+    $lines /= 2;
+    my $yt = int($y + $self->{h} / 2 - $em * $lines - $em * 0.1);
 
+    $yt += $self->{h} * 0.3 if $s =~ /^(triangle|trapezium)\z/;
+    $yt -= $self->{h} * 0.3 if $s =~ /^inv(triangle|trapezium)\z/;
+  
     my $color = $self->attribute('color') || 'black';
 
-    $svg .= "$indent<text class=\"text\" x=\"$xt\" y=\"$yt\" fill=\"$color\">$label</text>\n\n";
+    $svg .= $self->_svg_text($label, $color, $em, $indent, $xt, $yt);
     }
 
   $svg;
@@ -441,7 +609,12 @@ sub _svg_attributes
   my $y2 = $y + $self->{h} - $sub;
   $x += $sub; $y += $sub;
 
-  if ($shape eq 'point')
+  my $sub3 = $sub / 3;		# 0.333 * $sub
+  my $sub6 = 2 * $sub / 3;	# 0.666 * $sub
+
+  my $form = '%0.4f';		# limit precision to sane values
+
+  if ($shape =~ /^(point|none)\z/)
     {
     }
   elsif ($shape eq 'circle')
@@ -454,31 +627,31 @@ sub _svg_attributes
     }
   elsif ($shape eq 'parallelogram')
     {
-    my $xll = sprintf("%0.4f", $x - $sub * 0.333 + $self->{w} * 0.25);
-    my $xrl = sprintf("%0.4f", $x2 + $sub * 0.333 - $self->{w} * 0.25);
+    my $xll = sprintf($form, $x - $sub3 + $self->{w} * 0.25);
+    my $xrl = sprintf($form, $x2 + $sub3 - $self->{w} * 0.25);
 
-    my $xl = sprintf("%0.4f", $x + $sub * 0.666);
-    my $xr = sprintf("%0.4f", $x2 - $sub * 0.666);
+    my $xl = sprintf($form, $x + $sub6);
+    my $xr = sprintf($form, $x2 - $sub6);
 
     $shape = "polygon points=\"$xll,$y, $xr,$y, $xrl,$y2, $xl,$y2\"";
     }
   elsif ($shape eq 'trapezium')
     {
-    my $xl = sprintf("%0.4f", $x - $sub * 0.333 + $self->{w} * 0.25);
-    my $xr = sprintf("%0.4f", $x2 + $sub * 0.333 - $self->{w} * 0.25);
+    my $xl = sprintf($form, $x - $sub3 + $self->{w} * 0.25);
+    my $xr = sprintf($form, $x2 + $sub3 - $self->{w} * 0.25);
 
-    my $xl1 = sprintf("%0.4f", $x + $sub * 0.333);
-    my $xr1 = sprintf("%0.4f", $x2 - $sub * 0.333);
+    my $xl1 = sprintf($form, $x + $sub3);
+    my $xr1 = sprintf($form, $x2 - $sub3);
 
     $shape = "polygon points=\"$xl,$y, $xr,$y, $xr1,$y2, $xl1,$y2\"";
     }
   elsif ($shape eq 'invtrapezium')
     {
-    my $xl = sprintf("%0.4f", $x - $sub * 0.333 + $self->{w} * 0.25);
-    my $xr = sprintf("%0.4f", $x2 + $sub * 0.333 - $self->{w} * 0.25);
+    my $xl = sprintf($form, $x - $sub3 + $self->{w} * 0.25);
+    my $xr = sprintf($form, $x2 + $sub3 - $self->{w} * 0.25);
 
-    my $xl1 = sprintf("%0.4f", $x + $sub * 0.333);
-    my $xr1 = sprintf("%0.4f", $x2 - $sub * 0.333);
+    my $xl1 = sprintf($form, $x + $sub3);
+    my $xr1 = sprintf($form, $x2 - $sub3);
 
     $shape = "polygon points=\"$xl1,$y, $xr1,$y, $xr,$y2, $xl,$y2\"";
     }
@@ -487,34 +660,34 @@ sub _svg_attributes
     my $x1 = $cx;
     my $y1 = $cy;
 
-    my $xl = sprintf("%0.4f", $x + ($sub * 0.333));
-    my $xr = sprintf("%0.4f", $x2 - ($sub * 0.333));
+    my $xl = sprintf($form, $x + $sub3);
+    my $xr = sprintf($form, $x2 - $sub3);
 
     $shape = "polygon points=\"$xl,$y1, $x1,$y, $xr,$y1, $x1,$y2\" stroke-linecap=\"round\"";
     }
   elsif ($shape eq 'house')
     {
     my $x1 = $cx;
-    my $y1 = sprintf("%0.4f", $y - ($sub * 0.333) + $self->{h} * 0.333);
+    my $y1 = sprintf($form, $y - $sub3 + $self->{h} * 0.333);
 
     $shape = "polygon points=\"$x1,$y, $x2,$y1, $x2,$y2, $x,$y2, $x,$y1\"";
     }
   elsif ($shape eq 'pentagon')
     {
     my $x1 = $cx;
-    my $x11 = sprintf("%0.4f", $x - 0.333 * $sub + $self->{w} * 0.25);
-    my $x12 = sprintf("%0.4f", $x2 + 0.333 * $sub - $self->{w} * 0.25);
-    my $y1 = sprintf("%0.4f", $y - 0.666 * $sub + $self->{h} * 0.333);
+    my $x11 = sprintf($form, $x - $sub3 + $self->{w} * 0.25);
+    my $x12 = sprintf($form, $x2 + $sub3 - $self->{w} * 0.25);
+    my $y1 = sprintf($form, $y - $sub6 + $self->{h} * 0.333);
     
-    my $xl = sprintf("%0.4f", $x + 0.333 * $sub);
-    my $xr = sprintf("%0.4f", $x2 - 0.333 * $sub);
+    my $xl = sprintf($form, $x + $sub3);
+    my $xr = sprintf($form, $x2 - $sub3);
 
     $shape = "polygon points=\"$x1,$y, $xr,$y1, $x12,$y2, $x11,$y2, $xl,$y1\"";
     }
   elsif ($shape eq 'invhouse')
     {
     my $x1 = $cx;
-    my $y1 = sprintf("%0.4f", $y - (1.4 * $sub) + $self->{h} * 0.666);
+    my $y1 = sprintf($form, $y - (1.4 * $sub) + $self->{h} * 0.666);
 
     $shape = "polygon points=\"$x,$y, $x2,$y, $x2,$y1, $x1,$y2, $x,$y1\"";
     }
@@ -522,40 +695,40 @@ sub _svg_attributes
     {
     my $x15 = $cx;
     
-    my $x11 = sprintf("%0.4f", $x2 + 0.333 * $sub - $self->{w} * 0.10);
-    my $x14 = sprintf("%0.4f", $x - 0.333 * $sub + $self->{w} * 0.10);
+    my $x11 = sprintf($form, $x2 + $sub3 - $self->{w} * 0.10);
+    my $x14 = sprintf($form, $x - $sub3 + $self->{w} * 0.10);
 
-    my $y11 = sprintf("%0.4f", $y - 0.333 * $sub + $self->{h} * 0.15);
-    my $y13 = sprintf("%0.4f", $y2 + 0.85 * $sub - $self->{h} * 0.40);
+    my $y11 = sprintf($form, $y - $sub3 + $self->{h} * 0.15);
+    my $y13 = sprintf($form, $y2 + 0.85 * $sub - $self->{h} * 0.40);
 
-    my $x12 = sprintf("%0.4f", $x2 + $sub * 0.666 - $self->{w} * 0.25);
-    my $x13 = sprintf("%0.4f", $x - $sub * 0.666 + $self->{w} * 0.25);
+    my $x12 = sprintf($form, $x2 + $sub6 - $self->{w} * 0.25);
+    my $x13 = sprintf($form, $x - $sub6 + $self->{w} * 0.25);
     
-    my $xl = sprintf("%0.4f", $x - 0.15 * $sub);
-    my $xr = sprintf("%0.4f", $x2 + 0.15 * $sub);
+    my $xl = sprintf($form, $x - 0.15 * $sub);
+    my $xr = sprintf($form, $x2 + 0.15 * $sub);
 
     $shape = "polygon points=\"$x15,$y, $x11,$y11, $xr,$y13, $x12,$y2, $x13,$y2, $xl,$y13, $x14, $y11\"";
     }
   elsif ($shape eq 'octagon')
     {
-    my $x11 = sprintf("%0.4f", $x - 0.333 * $sub + $self->{w} * 0.25);
-    my $x12 = sprintf("%0.4f", $x2 + 0.333 * $sub - $self->{w} * 0.25);
-    my $y11 = sprintf("%0.4f", $y - 0.666 * $sub + $self->{h} * 0.25);
-    my $y12 = sprintf("%0.4f", $y2 + 0.666 * $sub - $self->{h} * 0.25);
+    my $x11 = sprintf($form, $x - $sub3 + $self->{w} * 0.25);
+    my $x12 = sprintf($form, $x2 + $sub3 - $self->{w} * 0.25);
+    my $y11 = sprintf($form, $y - $sub6 + $self->{h} * 0.25);
+    my $y12 = sprintf($form, $y2 + $sub6 - $self->{h} * 0.25);
 
-    my $xl = sprintf("%0.4f", $x + $sub * 0.133);
-    my $xr = sprintf("%0.4f", $x2 - $sub * 0.133);
+    my $xl = sprintf($form, $x + $sub * 0.133);
+    my $xr = sprintf($form, $x2 - $sub * 0.133);
 
     $shape = "polygon points=\"$xl,$y11, $x11,$y, $x12,$y, $xr,$y11, $xr,$y12, $x12,$y2, $x11,$y2, $xl,$y12\"";
     }
   elsif ($shape eq 'hexagon')
     {
     my $y1 = $cy;
-    my $x11 = sprintf("%0.4f", $x - 0.666 * $sub + $self->{w} * 0.25);
-    my $x12 = sprintf("%0.4f", $x2 + 0.666 * $sub - $self->{w} * 0.25);
+    my $x11 = sprintf($form, $x - $sub6 + $self->{w} * 0.25);
+    my $x12 = sprintf($form, $x2 + $sub6 - $self->{w} * 0.25);
 
-    my $xl = sprintf("%0.4f", $x + $sub * 0.333);
-    my $xr = sprintf("%0.4f", $x2 - $sub * 0.333);
+    my $xl = sprintf($form, $x + $sub3);
+    my $xr = sprintf($form, $x2 - $sub3);
 
     $shape = "polygon points=\"$xl,$y1, $x11,$y, $x12,$y, $xr,$y1, $x12,$y2, $x11,$y2\"";
     }
@@ -563,10 +736,10 @@ sub _svg_attributes
     {
     my $x1 = $cx;
 
-    my $xl = sprintf("%0.4f", $x + $sub);
-    my $xr = sprintf("%0.4f", $x2 - $sub);
+    my $xl = sprintf($form, $x + $sub);
+    my $xr = sprintf($form, $x2 - $sub);
 
-    my $yd = sprintf("%0.4f", $y2 + ($sub * 0.2 ));
+    my $yd = sprintf($form, $y2 + ($sub * 0.2 ));
 
     $shape = "polygon points=\"$x1,$y, $xr,$yd, $xl,$yd\"";
     }
@@ -574,10 +747,10 @@ sub _svg_attributes
     {
     my $x1 = $cx;
 
-    my $xl = sprintf("%0.4f", $x + $sub);
-    my $xr = sprintf("%0.4f", $x2 - $sub);
+    my $xl = sprintf($form, $x + $sub);
+    my $xr = sprintf($form, $x2 - $sub);
 
-    my $yd = sprintf("%0.4f", $y - ($sub * 0.2 ));
+    my $yd = sprintf($form, $y - ($sub * 0.2));
 
     $shape = "polygon points=\"$xl,$yd, $xr,$yd, $x1,$y2\"";
     }
@@ -592,9 +765,9 @@ sub _svg_attributes
     {
     if ($shape eq 'rounded')
       {
-      # round corners
-      $att->{ry} = '15%';
-      $att->{rx} = '15%';
+      # round corners by a fixed value
+      $att->{ry} = '15';
+      $att->{rx} = '15';
       $shape = 'rect';
       }
     $att->{x} = $x;
@@ -624,6 +797,7 @@ sub _svg_attributes
     delete $att->{stroke};
     }
   $att->{fill} = $self->attribute('fill') || 'white';
+  delete $att->{fill} if $att->{fill} eq 'white';	# white is default
 
   $att;
   }
@@ -638,7 +812,7 @@ sub _svg_attributes_as_txt
   foreach my $e (sort keys %$att)
     {
     # skip these
-    next if $e =~ /^(arrow-style|label-color|rows|cols|size|offset|origin)\z/;
+    next if $e =~ /^(arrow-style|text-style|label-color|rows|cols|size|offset|origin)\z/;
 
     $att_line .= " $e=\"$att->{$e}\"";
     if (length($att_line) > 75)
@@ -647,7 +821,7 @@ sub _svg_attributes_as_txt
       }
     }
   $att_txt .= $att_line;
-  $att_txt =~ s/\n\z//;		# avoid a "  >" on last line
+  $att_txt =~ s/\n  \z//;		# avoid a "  >" on last line
   $att_txt;
   }
  
@@ -656,7 +830,7 @@ sub _correct_size_svg
   # Correct {w} and {h} after parsing.
   my $self = shift;
 
-  my $em = Graph::Easy::EM;		# multiplication factor chars * em = units (pixels)
+  my $em = $self->EM();		# multiplication factor chars * em = units (pixels)
 
   return if defined $self->{w};
 
@@ -668,28 +842,27 @@ sub _correct_size_svg
     return;
     }
 
-  #my $txt = $self->label();
-
   my ($w,$h) = $self->dimensions();
 
   # XXX TODO: that should use a changable padding factor (like "0.2 em" or "4")
-  $self->{w} = int($w * $em + 0.2 * $em);
+  $self->{w} = int($w * $em + 0.3 * $em);
   $self->{h} = int($h * $em + 0.8 * $em);
 
   my $border = $self->attribute('border-style') || 'none';
+  $border = 'none' if $shape eq 'none';
   if ($border ne 'none')
     {
-    # XXX TODO: that should use the border width (like "1 pixel" * 2)
-    $self->{w} += 2;
-    $self->{h} += 2;
+    my $bw = $self->attribute('border-width') || '1';
+    $self->{w} += $bw * 2;
+    $self->{h} += $bw * 2;
     }
  
   # for triangle or invtriangle:
-  $self->{w} *= 1.5 if $shape =~ /triangle/;
-  $self->{h} *= 2;# if $shape =~ /triangle|trapezium/;
+  $self->{w} *= 1.4 if $shape =~ /triangle/;
+  $self->{h} *= 1.8 if $shape =~ /triangle|trapezium/;
   $self->{w} *= 1.2 if $shape =~ /(parallelogram|trapezium|pentagon)/;
 
-  if ($shape =~ /^(diamond|circle|octagon|hexagon)\z/)
+  if ($shape =~ /^(diamond|circle|octagon|hexagon|triangle)\z/)
     {
     # the min size is either w or h, depending on which is bigger
     my $max = $self->{w}; $max = $self->{h} if $self->{h} > $max;
@@ -762,7 +935,7 @@ sub _svg_arrow
   my $a = {};
   for my $key (keys %$att)
     {
-    next if $key =~ /^(stroke-dasharray|arrow-style)\z/;
+    next if $key =~ /^(stroke-dasharray|arrow-style|stroke-width)\z/;
     $a->{$key} = $att->{$key}
      unless exists $DEF->{$key} && $DEF->{$key} eq $att->{$key};
     }
@@ -779,6 +952,8 @@ sub _svg_arrow
 
   my $att_txt = $self->_svg_attributes_as_txt($a);
 
+  $self->{graph}->_svg_use_def($class) if ref $self->{graph};
+
   my $ar = "$indent<use$att_txt xlink:href=\"#$class\" ";
 
   my $svg = '';
@@ -826,13 +1001,13 @@ sub _svg_line_straight
   # "$s" means there is a starting point, so the line needs to be shorter. Likewise
   # for "$e", only on the "other" side. 
   # VER: s = north, e = south, HOR: s = left, e= right
-  my ($self, $att, $x, $y, $type, $l, $r, $indent, $s, $e) = @_;
-
-  my $att_txt = $self->_svg_attributes_as_txt($att);
+  my ($self, $x, $y, $type, $l, $r, $s, $e, $add) = @_;
 
   my $w = $self->{w};
   my $h = $self->{h};
 
+  $add = '' unless defined $add;	# additinal styles?
+
   my ($x1,$x2, $y1,$y2, $x3, $x4, $y3, $y4);
 
   my $ltype = $type & LINE_MASK;
@@ -847,7 +1022,8 @@ sub _svg_line_straight
     if (($type & LINE_DOUBLE) != 0)
       {
       $y1--; $y2--; $y3 = $y1 + 2; $y4 = $y3;
-      $x1 += 1.5; $x2 -= 1.5;
+      # shorten the line for end/start points
+      if ($s || $e ) { $x1 += 1.5; $x2 -= 1.5; }
       $x3 = $x1; $x4 = $x2;
       }
     }
@@ -862,48 +1038,46 @@ sub _svg_line_straight
     if (($type & LINE_DOUBLE) != 0)
       {
       $x1--; $x2--; $x3 = $x1 + 2; $x4 = $x3;
-      $y1 += 1.5; $y2 -= 1.5;
+      # shorten the line for end/start points
+      if ($s || $e) { $y1 += 1.5; $y2 -= 1.5; }
       $y3 = $y1; $y4 = $y2;
       }
     }
 
-  my $txt = 
-  "$indent<line x1=\"$x1\" y1=\"$y1\" x2=\"$x2\" y2=\"$y2\"$att_txt />\n";
+  my @r = ( "<line x1=\"$x1\" y1=\"$y1\" x2=\"$x2\" y2=\"$y2\" $add/>" );
 
   # for a double line
-  if (defined $x3)
-    {
-    $txt .=
-     "$indent<line x1=\"$x3\" y1=\"$y3\" x2=\"$x4\" y2=\"$y4\"$att_txt />\n";
-    }
-  $txt;
+  push @r, "<line x1=\"$x3\" y1=\"$y3\" x2=\"$x4\" y2=\"$y4\" $add/>"
+   if defined $x3;
+
+  @r;
   }
 
 #############################################################################
 #############################################################################
 
 my $dimensions = {
-  EDGE_VER()	=> [ 1, 4 ],	# |
-  EDGE_HOR()	=> [ 4, 1 ],	# -
+  EDGE_VER()	=> [ 1, 3 ],	# |
+  EDGE_HOR()	=> [ 3, 1 ],	# -
 
-  EDGE_CROSS()	=> [ 4, 4 ],	# +	crossing
+  EDGE_CROSS()	=> [ 3, 3 ],	# +	crossing
 
-  EDGE_N_E()	=> [ 4, 4 ],	# |_    corner (N to E)
-  EDGE_N_W()	=> [ 4, 4 ],	# _|    corner (N to W)
-  EDGE_S_E()	=> [ 4, 4 ],	# ,-    corner (S to E)
-  EDGE_S_W()	=> [ 4, 4 ],	# -,    corner (S to W)
+  EDGE_N_E()	=> [ 3, 3 ],	# |_    corner (N to E)
+  EDGE_N_W()	=> [ 3, 3 ],	# _|    corner (N to W)
+  EDGE_S_E()	=> [ 3, 3 ],	# ,-    corner (S to E)
+  EDGE_S_W()	=> [ 3, 3 ],	# -,    corner (S to W)
 
-  EDGE_S_E_W	=> [ 4, 2 ],	# -,-   three-sided corner (S to W/E)
-  EDGE_N_E_W	=> [ 4, 2 ],	# -'-   three-sided corner (N to W/E)
-  EDGE_E_N_S	=> [ 2, 4 ],	#  |-   three-sided corner (E to S/N)
-  EDGE_W_N_S	=> [ 2, 4 ], 	# -|    three-sided corner (W to S/N)
+  EDGE_S_E_W	=> [ 3, 3 ],	# -,-   three-sided corner (S to W/E)
+  EDGE_N_E_W	=> [ 3, 3 ],	# -'-   three-sided corner (N to W/E)
+  EDGE_E_N_S	=> [ 3, 3 ],	#  |-   three-sided corner (E to S/N)
+  EDGE_W_N_S	=> [ 3, 3 ], 	# -|    three-sided corner (W to S/N)
  };
 
 sub _correct_size_svg
   {
   my ($self,$format) = @_;
 
-  my $em = Graph::Easy::EM;		# multiplication factor chars * em = units (pixels)
+  my $em = $self->EM();		# multiplication factor chars * em = units (pixels)
 
   return if defined $self->{w};
 
@@ -914,12 +1088,14 @@ sub _correct_size_svg
   my $dim = $dimensions->{$type} || [ 2, 2 ];
   ($self->{w}, $self->{h}) = ($dim->[0], $dim->[1]);
 
-  # XXX TODO
-  my @lines = $self->_formatted_label();
-  if (@lines > 1)
+  # make it bigger for cells with the label
+  if ($self->{type} & EDGE_LABEL_CELL)
     {
-    $self->{h} = @lines + 3;
+    my ($w,$h) = $self->dimensions();
+    $self->{w} += $w - 2;
+    $self->{h} += $h;
     }
+
   ($self->{w}, $self->{h}) = ($self->{w} * $em, $self->{h} * $em);
   }
 
@@ -934,6 +1110,8 @@ sub _svg_attributes
   my $att = {};
 
   $att->{stroke} = $self->attribute('color') || 'black';
+  delete $att->{stroke} if $att->{stroke} eq 'black';	# black is default
+
   $att->{'stroke-width'} = 1;
 
   my $style = $self->{style};
@@ -953,11 +1131,11 @@ sub as_svg
   {
   my ($self,$x,$y, $indent) = @_;
 
-  my $em = Graph::Easy::EM;		# multiplication factor chars * em = units (pixels)
+  my $em = $self->EM();		# multiplication factor chars * em = units (pixels)
 
   # the attributes of the element we will finally output
   my $att = $self->_svg_attributes();
-  
+ 
   # set a potential title
   my $title = $self->title();
   $att->{title} = $title if $title ne '';
@@ -973,12 +1151,18 @@ sub as_svg
   my $to = $edge->{to}->{name};
   my $svg = "$indent<!-- " . edge_type($type) . ", from $from to $to -->\n";
 
-  $self->_svg_background(\$svg, $x,$y, $indent);# if defined $att->{background} || keys %{$self->{edge}->{groups}} > 0;
-
-  # for each line, include a SVG tag
+  $self->_svg_background(\$svg, $x,$y, $indent);
+  # if defined $att->{background} || keys %{$self->{edge}->{groups}} > 0;
 
+  # for each line, include one SVG tag
   my $lines = [ @{$draw_lines->{$type}} ];	# make copy
 
+  my $style = $self->{style};
+
+  my $cross = ($self->{type} & EDGE_TYPE_MASK) == EDGE_CROSS;	# we are a cross section?
+  my $add;
+
+  my @line_tags;
   while (@$lines > 0)
     {
     my ($type, $l, $r) = splice (@$lines, 0, 3);
@@ -1002,10 +1186,35 @@ sub as_svg
       $s = 0.1 if ($start & EDGE_START_W);
       }
 
-    $type += LINE_DOUBLE if $self->{style} =~ /^double/;
-    $svg .= $self->_svg_line_straight($att, $x, $y, $type, $l, $r, $indent, $s, $e);
+    # LINE_VER must come last
+    if ($cross && $type == LINE_VER)
+      {
+      $style = $self->{style_ver};
+      $add = ' stroke="' . $self->{color_ver} . '"' if $self->{color_ver};
+      $add .= ' stroke-dasharray="' . ($strokes->{$style}||'1 0') .'"';
+      }
+    $type += LINE_DOUBLE if $style =~ /^double/;
+    push @line_tags, $self->_svg_line_straight($x, $y, $type, $l, $r, $s, $e, $add);
     }
 
+  # we can put the line tags into a <g> and put stroke attributes on the g,
+  # this will shorten the output
+
+  $lines = ''; my $p = "\n"; my $i = $indent;
+  if (@line_tags > 1)
+    {
+    $lines = "$indent<g$att_txt>\n";
+    $i .= $indent;
+    $p = "\n$indent</g>\n"; 
+    }
+  else
+    {
+    $line_tags[0] =~ s/ \/>/$att_txt \/>/;
+    }
+  $lines .= $i . join("\n$i", @line_tags) . $p;
+
+  $svg .= $lines;
+
   my $arrow = $self->{type} & EDGE_END_MASK;
   # depending on end points, add the arrow
 
@@ -1013,32 +1222,27 @@ sub as_svg
 
   ###########################################################################
   # include the label/name/text if we are the label cell
- 
+
   if (($self->{type} & EDGE_LABEL_CELL()))
     {
     my $label = $self->label(); $label = '' unless defined $label;
 
     if ($label ne '')
       {
-      my $xt = int($x + $em * 0.2 + $self->{w} / 4);
+      my $xt = int($x + $em * 0.2 + $self->{w} / 2);
       my $yt = int($y + $em + $em * 0.2);
 
-      # XXX TODO: use attribute "label-color"
-      my $color = $self->attribute('label-color') || $att->{color} || 'black';
+      my $color = $self->attribute('label-color') || '';
 
-      $label =~ s/\s*\\n\s*/\n/g;			# insert real newlines
+      # fall back to color if label-color not defined
+      $color = $self->attribute('color') || 'black' if $color eq '';
 
-      my @lines = split/\n/,$label;			# split into lines
-
-      if (@lines > 1)
-        {
-        $label = "<tspan>" . join ('</tspan><tspan dy="1em">', @lines) . '</tspan>';
-        }
-
-      $svg .= "$indent<text x=\"$xt\" y=\"$yt\" style=\"font-size:$em\" fill=\"$color\">$label</text>\n";
+      $svg .= $self->_svg_text($label, $color, $em, $indent, $xt, $yt);
       }
-    }
+    } 
 
-  $svg. "\n";
+  $svg .= "\n" unless $svg =~ /\n\n\z/;
+
+  $svg;
   }
 
